/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::SkillManager
 * @author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <RobotAPI/interface/skills/SkillProviderInterface.ice>
#include <RobotAPI/interface/aron.ice>

module armarx
{
    module skills
    {
        module manager
        {
            module dto
            {
                struct ProviderID
                {
                    string providerName;
                };

                // A skill ID. Must be unique within one SkillManager
                struct SkillID
                {
                    ProviderID providerId;
                    string skillName;
                };

                struct SkillDescription
                {
                    SkillID skillId;
                    string description;
                    armarx::core::time::dto::Duration timeout;
                    aron::type::dto::AronObject parametersType;
                    aron::type::dto::AronObject resultType;
                    aron::data::dto::Dict rootProfileDefaults;
                };

                dictionary<SkillID, dto::SkillDescription> SkillDescriptionMap;

                struct ProviderInfo
                {
                    ProviderID providerId;
                    provider::dti::SkillProviderInterface* providerInterface;
                    dto::SkillDescriptionMap providedSkills;
                };

                struct SkillExecutionID
                {
                    SkillID skillId;
                    string executorName;
                    armarx::core::time::dto::DateTime executionStartedTime;
                    string uuid;
                };

                struct SkillStatusUpdate
                {
                    SkillExecutionID executionId;
                    aron::data::dto::Dict parameters;
                    callback::dti::SkillProviderCallbackInterface* callbackInterface;
                    core::dto::Execution::Status status;
                    aron::data::dto::Dict result;
                };

                dictionary<SkillExecutionID, SkillStatusUpdate> SkillStatusUpdateMap;

                struct SkillExecutionRequest
                {
                    SkillID skillId;
                    string executorName;
                    aron::data::dto::Dict parameters;
                };

                //****************************//
                //** Fluxio related structs **//
                //****************************//

                struct FluxioIdentificator
                {
                    string id;
                    string hint;
                };

                struct FluxioTypeIdentificator
                {
                  string skillId;
                  string parameterId;
                };

                exception FluxioException  
                {
                    string message;
                    string type;
                    Ice::StringSeq contextList;
                };

                sequence<dto::FluxioIdentificator> FluxioIdentificatorList;

                struct FluxioSkillStatusUpdate
                {
                    armarx::core::time::dto::DateTime timestamp;
                    string executionId;
                    FluxioIdentificator subSkillNodeId;
                    core::dto::Execution::Status status;
                };

                sequence<FluxioSkillStatusUpdate> FluxioSkillStatusUpdateList;

                struct FluxioProfile
                {
                    string id;
                    string name;
                    string description;
                    bool parentIdHasValue;
                    FluxioIdentificator parentId;
                };
                
                sequence<FluxioProfile> FluxioProfileList;

                struct FluxioProvider
                {
                    string id;
                    string name;
                };

                sequence<FluxioProvider> FluxioProviderList;

                struct FluxioValue
                {
                    FluxioIdentificator profileId;
                    aron::data::dto::GenericData content;
                };

                sequence<FluxioValue> FluxioValueList;

                struct FluxioParameter
                {
                    string id;
                    string name;
                    string description;
                    FluxioTypeIdentificator type;
                    bool required;
                    bool isInput;
                    FluxioValueList values;
                };

                struct FluxioEdge
                {
                    FluxioIdentificator fromNodeId;
                    FluxioIdentificator fromParameterId;
                    FluxioIdentificator toNodeId;
                    FluxioIdentificator toParameterId;
                };

                sequence<FluxioParameter> FluxioParameterList;
                
                struct FluxioNode
                {
                    string nodeType; // decides which extended attributes are used
                    string nodeId;
                    string name;
                    float xPos;
                    float yPos;
                    FluxioIdentificator parameterId; // empty id field if node is not a parameter node
                    FluxioIdentificator skillId; // empty id field if node is not a subskill node   
                    string controlType; // empty string if node is not a control node
                    FluxioParameterList parameters; // empty list if node is not a control node                     
                };

                sequence<FluxioNode> FluxioNodeList;
                sequence<FluxioEdge> FluxioEdgeList;

                struct FluxioSkill
                {
                    string id;
                    string name;
                    string description;
                    armarx::core::time::dto::Duration timeout;
                    string lastChanged;
                    bool executable;
                    bool native;
                    FluxioIdentificator skillProviderId;
                    FluxioParameterList parameters;
                    bool nodesHasValue;
                    FluxioNodeList nodes;
                    bool edgesHasValue;
                    FluxioEdgeList edges;
                };

                sequence<dto::FluxioSkill> FluxioSkillList;
            }

            module dti
            {
                sequence<provider::dto::AbortSkillResult> AbortSkillResultList;

                interface SkillManagerInterface extends
                    callback::dti::SkillProviderCallbackInterface
                {
                    // !! There is by design no method to get a proxy of the provider. You should only communicate with the manager !!

                    void 
                    addProvider(dto::ProviderInfo providerInfo);

                    void 
                    removeProvider(dto::ProviderID providerId);

                    optional(1) dto::SkillDescription 
                    getSkillDescription(dto::SkillID skillId); // get one skilldescriptions from a provider

                    dto::SkillDescriptionMap
                    getSkillDescriptions(); // get all skilldescriptions from all providers

                    optional(2) dto::SkillStatusUpdate
                    getSkillExecutionStatus(dto::SkillExecutionID executionId);

                    dto::SkillStatusUpdateMap
                    getSkillExecutionStatuses(); // returns the current executions from all providers

                    dto::SkillStatusUpdate
                    executeSkill(dto::SkillExecutionRequest skillExecutionRequest); // blocks until skill is finished.

                    dto::SkillExecutionID 
                    executeSkillAsync(dto::SkillExecutionRequest skillExecutionRequest); // directly returns the execution id

                    provider::dto::ParameterUpdateResult 
                    updateSkillParameters(dto::SkillExecutionID executionId, aron::data::dto::Dict parameters); // add params to a skill

                    // notify a skill to stop ASAP.
                    provider::dto::AbortSkillResult 
                    abortSkill(dto::SkillExecutionID executionId); 

                    provider::dto::AbortSkillResult
                    abortSkillAsync(dto::SkillExecutionID executionId);

                    // notify all skills directly via ice interface instead of ugly workarounds in both c++ and python
                    AbortSkillResultList abortAllSkills();
                    AbortSkillResultList abortAllSkillsAsync();

                    //****************************//
                    //** Fluxio related methods **//
                    //****************************//

                    aron::type::dto::AronObject
                    getTypes();

                    optional(3) string
                    executeFluxioSkill(string skillId, string profileId) throws dto::FluxioException; // executes a fluxio skill
                    
                    void
                    abortFluxioSkill(string executionId) throws dto::FluxioException ; // aborts a fluxio skill

                    dto::FluxioSkillList
                    getSkillList() throws dto::FluxioException; // returns skills of all providers
                    
                    optional(4) dto::FluxioSkill 
                    getSkill(string id) throws dto::FluxioException; // returns a single skill

                    bool 
                    updateSkill(string userId, dto::FluxioSkill skill) throws dto::FluxioException; // updates an existing skill

                    bool
                    updateSkillParameterValues(string userId, string skillId, string parameterId, dto::FluxioValueList values) throws dto::FluxioException; // updates the profile values of a skill parameter

                    bool
                    updateSkillValues(string userId, string skillId, dto::FluxioParameterList parameters) throws dto::FluxioException; // updates the profile values of a skill parameter

                    bool 
                    getSkillMutex(string skillId, string userId) throws dto::FluxioException; // request a mutex for a skill

                    void 
                    deleteSkillMutex(string skillId, string userId) throws dto::FluxioException; // release a mutex for a skill

                    optional(5) dto::FluxioIdentificatorList
                    deleteSkill(string skillId, string userId, bool dryRun) throws dto::FluxioException; // removes a skill

                    optional(6) dto::FluxioIdentificatorList
                    deleteSkillParameter(string skillId, string parameterId, string userId, bool dryRun) throws dto::FluxioException; // deletes a parameter from a skill
                    
                    optional(7) dto::FluxioIdentificatorList
                    updateSkillParameter(string skillId, dto::FluxioParameter parameter, string userId, bool dryRun) throws dto::FluxioException; // updates a parameter of a skill

                    dto::FluxioProfileList
                    getProfileList() throws dto::FluxioException; // returns all profiles

                    optional(8) dto::FluxioProfile
                    getProfile(string id) throws dto::FluxioException; // returns a single profile

                    dto::FluxioProfile 
                    createProfile(dto::FluxioProfile profile) throws dto::FluxioException; // creates a new profile

                    void 
                    updateProfile(dto::FluxioProfile profile) throws dto::FluxioException; // updates an existing profile

                    dto::FluxioProviderList
                    getProviderList() throws dto::FluxioException; // returns all providers

                    optional(9) dto::FluxioProvider
                    getProvider(string id) throws dto::FluxioException; // returns a single provider

                    optional(10) dto::FluxioSkillList
                    getSkillsOfProvider(string id) throws dto::FluxioException; // returns all skills of a provider

                    optional(11) dto::FluxioSkill
                    addSkillToProvider(string userId, string providerId, dto::FluxioSkill skill) throws dto::FluxioException; // adds a new skill to a provider

                    optional(12) dto::FluxioSkillStatusUpdateList
                    getFluxioSkillExecutionStatus(string executionId) throws dto::FluxioException; // returns the status of a fluxio execution
                };
            }
        }
    }
}
