/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <string>

#include <VirtualRobot/VirtualRobot.h>
#include <VirtualRobot/XML/RobotIO.h>

#include <RobotAPI/libraries/armem/core/forward_declarations.h>

#include "RobotReader.h"

namespace armarx::armem::robot_state
{
    /**
     * @brief The VirtualRobotReader class.
     *robot
     * The aim of this class is to obtain a virtual robot instance and synchronize it
     * with the data (joint positions, global pose, ...) stored in the working memory.
     *
     * This is only a lightweight wrapper of @see RobotReader for Simox's VirtualRobot class.
     */
    class VirtualRobotReader : virtual public RobotReader
    {
    public:
        using RobotReader::RobotReader;

        ~VirtualRobotReader() override = default;

        /**
         * @brief Synchronize both the platform pose and the joint values of a virtual robot,
         *        according to the robot state memory for a certain point in time.
         * @param robot whose platform pose and joint values to modify
         * @param timestamp for which to synchronize
         * @return whether synchronization was successful
         */
        [[nodiscard]] bool synchronizeRobot(VirtualRobot::Robot& robot,
                                            const armem::Time& timestamp) const;

        /**
         * @brief Synchronize only the joint values of a virtual robot,
         *        according to the robot state memory for a certain point in time.
         * @param robot whose joint values to modify
         * @param timestamp for which to synchronize
         * @return whether synchronization was successful
         */
        [[nodiscard]] bool synchronizeRobotJoints(VirtualRobot::Robot& robot,
                                                  const armem::Time& timestamp) const;

        /**
         * @brief Synchronize only the platform pose of a virtual robot,
         *        according to the robot state memory for a certain point in time.
         * @param robot whose platform pose and joint values to modify
         * @param timestamp for which to synchronize
         * @return whether synchronization was successful
         */
        [[nodiscard]] bool synchronizeRobotPose(VirtualRobot::Robot& robot,
                                                const armem::Time& timestamp) const;

        [[nodiscard]] VirtualRobot::RobotPtr
        getRobot(const std::string& name,
                 const armem::Time& timestamp = armem::Time::Invalid(),
                 const VirtualRobot::RobotIO::RobotDescription& loadMode =
                     VirtualRobot::RobotIO::RobotDescription::eStructure);

        [[deprecated(
            "Use getRobot() and synchronizeRobot() instead. This method does two things "
            "(getRobot() and synchronizeRobot()) and is not recommended as we cannot differentiate "
            "failure cases, e.g. we might be able to obtain the robot model but cannot synchronize "
            "it. Also, if `blocking` is enabled, this method will block indefinitely if the robot "
            "is not available.")]] 
        [[nodiscard]] VirtualRobot::RobotPtr
        getSynchronizedRobot(const std::string& name,
                             const VirtualRobot::RobotIO::RobotDescription& loadMode =
                                 VirtualRobot::RobotIO::RobotDescription::eStructure,
                             bool blocking = true);

        [[deprecated(
            "Use getRobot() and synchronizeRobot() instead. This method does two things "
            "(getRobot() and synchronizeRobot()) and is not recommended as we cannot differentiate "
            "failure cases, e.g. we might be able to obtain the robot model but cannot synchronize "
            "it. Also, if `blocking` is enabled, this method will block indefinitely if the robot "
            "is not available.")]] 
        [[nodiscard]] VirtualRobot::RobotPtr
        getSynchronizedRobot(const std::string& name,
                             const armem::Time& timestamp,
                             const VirtualRobot::RobotIO::RobotDescription& loadMode =
                                 VirtualRobot::RobotIO::RobotDescription::eStructure,
                             bool blocking = true);

        /**
        @brief In contrast to getRobot(), this function will retry to query the robot until it exists.
        */
        [[nodiscard]] VirtualRobot::RobotPtr
        getRobotWaiting(const std::string& name,
                        const armem::Time& timestamp = armem::Time::Invalid(),
                        const VirtualRobot::RobotIO::RobotDescription& loadMode =
                            VirtualRobot::RobotIO::RobotDescription::eStructure);

    private:
        [[nodiscard]] VirtualRobot::RobotPtr
        _getSynchronizedRobot(const std::string& name,
                              const armem::Time& timestamp = armem::Time::Invalid(),
                              const VirtualRobot::RobotIO::RobotDescription& loadMode =
                                  VirtualRobot::RobotIO::RobotDescription::eStructure,
                              bool blocking = true);
    };

} // namespace armarx::armem::robot_state
